/*
 * AMPL interface for E04UGC (SQP for sparse nonlinear programming problems)
 * intended for self-building by NAG C Library users
 * (Windows & Linux)
 *
 * version: 0.9 (beta)
 *
 * last update: 29/11/2010
 * by Jan Fiala, jan@nag.co.uk
 *
 *
 * Instructions:
 * (1) Prerequsities
 *     * ASL (AMPL Solver Library)
 *     * NAG C Library
 *
 *   If you don't have a suitable build of the ASL library you can
 *   download its source codes from
 *       http://netlib.sandia.gov/cgi-bin/netlib/netlibfiles.tar?filename=netlib/ampl/solvers
 *   as a tar archive. Or from
 *       http://www.netlib.org/ampl/solvers/
 *       http://netlib.sandia.gov/ampl/solvers/index.html
 *   as separate files. Please unzip it and follow README file for
 *   further instructions.
 *
 * (2) Building
 *   Choose appropriate Makefile according to your library and compiler.
 *   For example 
 *       Makefile.clw3209dal.cl
 *   is suitable if you use MS Visual C++ compiler (cl) and NAG C
 *   Library mark 9 for Windows (see comments inside). Edit library
 *   paths and other settings as desired. To build AMPL solvers invoke
 *   'make' on target 'all' as described inside the file depending on
 *   your system, for example:
 *       make -f Makefile.clw3209dal.gcc all
 *       nmake -f Makefile.clw3209dal.cl all
 *   If there is no Makefile for your specific combination of the system,
 *   library and compiler, edit a generic one as necessary.
 *
 * (3) Testing
 *   Invoke 'make' on target 'test' to run the embedded test, for example
 *       make -f Makefile.clw3209dal.gcc test
 *       nmake -f Makefile.clw3209dal.cl test
 *   By now your AMPL solver should work. In case of any troubles, please
 *   do not hesitate to contact NAG.
 *
 * Common problems:
 * (a) If you use dynamic link with the NAG Library don't forget to include
 * the library path to the system variables (PATH in Windows, LD_LIBRARY_PATH
 * in Linux). If there are any doubts if the library and/or licence is
 * accesible run the executable without any arguments. This will check the
 * licence and library calls. 
 * (b) When you call the solver from AMPL by "option solver solver_name;"
 * please do not forget that the solver is called directly by solver_name
 * thus 'solver_name' must include an absolute path or a relative path to
 * the current location. For example
 *     option solver e04ug;
 * (if you are on Windows in the same directory as the solver)
 *     option solver "./e04ug.exe";
 * (if you are on Linux in the same directory as the solver but the directory
 * is not included in your PATH variable)
 *
 *
 * NAG Lib calls: 
 *    nag_licence_query (a00acc)
 *    [nag_implementation_details (a00aac)]
 *    nag_opt_init (e04xxc)
 *    nag_opt_nlp_sparse (e04ugc)
 *    nag_opt_free (e04xzc)
 */

#include <stdio.h>
#include <malloc.h>

// AMPL interface library (their IO)
// [needs to go first because of NAG redefinitions of Vsprintf etc.]
#include "stdio1.h"

// declarations of NAG C Library subroutines, data types, error messages etc.
#include <nag.h>
#include <nag_stdlib.h>
#include <naga00.h>
#include <nage04.h>

// AMPL interface library (the rest)
#include "asl.h"
#include "getstub.h"

// verify derivatives? No, not necessary, are exact (otherwise comment it out)
#define DONOTVERIFY

// definition of keywords to modify default option settings of E04UGC
typedef struct {
  double tol_feas;   // Feasibility Tolerance
  double tol_prec;   // Function Precision
  double tol_optim;  // optimaliry tolerance
  double min_feas;   // Minor Feasibility Tolerance
  int maxit;         // Major Iteration Limit
  int minit;         // Minor Iteration Limit
  int minittotal;    // Iteration Limit (tummed minor iterations)
  int printlev;      // Print Level
  int mprintlev;     // minor print level
  int superbasics;   // Superbasics Limit
  int hesupdates;    // Hessian Updates
} e04ugc_parameters;

// callback functions declarations
#ifdef __cplusplus
extern "C" {
#endif
static void NAG_CALL confun(Integer ncnln, Integer njnln,
                            Integer nnzjac, const double x[], double conf[],
                            double conjac[], Nag_Comm *comm);

static void NAG_CALL objfun(Integer nonln,
                            const double x[], double *objf,
                            double objgrad[], Nag_Comm *comm);
#ifdef __cplusplus
}
#endif

// Ampl Solver Library (ASL) data structure
// (doesn't need to be global but it makes life easier...)
static ASL *asl=0;

// to store any deviations from default opt. settings, -1 to indicate no change
static e04ugc_parameters param = {-1, -1, -1, -1, -1, -1, -1, 2, -1, -1, -1}; 

// definition of keywords AMPL should react to
static keyword keywds[] = {
    KW("feasibility",D_val, &param.tol_feas,	"Major Feasibility Tolerance (default sqrt(eps))"),
    KW("hupdates", I_val, &param.hesupdates, "Hessian Updates (default 20 or Inf)"),
    KW("maxit",	   I_val, &param.maxit, "Major Iteration Limit (default 1000)"),
    KW("minfeas",D_val, &param.min_feas,	"Minor Feasibility Tolerance (default sqrt(eps))"),
    KW("minit",	     I_val, &param.minit, "Minor Iteration Limit (default 500)"),
    KW("minittotal", I_val, &param.minittotal, "Iteration Limit - summed minor iterations (default 10000)"),
    KW("minprint",   I_val, &param.mprintlev, "Minor Print Level (0 or 1, default 0)"),
    KW("optimality", D_val, &param.tol_optim,	"Optimality Tolerance (default sqrt(eps))"),
    KW("precision",	 D_val, &param.tol_prec,	"Function Precision (default eps^0.8)"),
    KW("print",      I_val, &param.printlev, "Print Level (0..4, default 2)"),
    KW("superbasics",I_val, &param.superbasics, "Superbasics Limit (default min(500,n+1))"),
  };

// usage message to be printed for '-?'
static char *usage_msg[] = {
  "  where  stub  is a file created from  `ampl -obstub`  or  `ampl -ogstub`",
  "  call.  Assignments have the form  `keyword=value`. To get a list of all",
  "  keywords, call `./e04ug.exe -=`, see description bellow.",
  "  Example:",
  "    ./e04ug.exe problem.nl maxit=100 precision=1e-2 print=0",
  " ",
  "  Alternatively, you can invoke solver E04UGC from Ampl environment. Run",
  "  `ampl` and input the following commands:",
  "     model problem.mod;",
  "     option solver e04ug;",
  "     option e04ug_options 'maxit=100 precision=1e-2 print=1';",
  "     solve;",
  0 };

static char *end_msg[] = {
  /*0*/ "converged: optimal solution found",
  /*1*/ "failed: the problem is infeasible",
  /*2*/ "failed: the problem seems to be unbounded",
  /*3*/ "error: error in the option settings",
  /*4*/ "failed: maximum number of iterations reached",
  /*5*/ "converged: suboptimal solution found, accuracy couldn't be achieved",
  /*6*/ "error: superbasics limit is too small, increase it using option 'superbasics'",
  /*7*/ "error: please consult the manual"
  };

// info structure of the solver, messages, option settings
static Option_Info Oinfo = {"E04UG", "E04UG, NAG C Library","e04ug_options", keywds, nkeywds, 0 /*funcadd?*/, "E04UG, NAG C Library, AMPL interface 0.9 (beta)", usage_msg, 0, 0, 0 /*cmd line options?*/, 0, 20101125};

// set non-default option settings given by user through AMPL
void set_options(Nag_E04_Opt *options) {

  const int prnlevels[]={Nag_NoPrint, Nag_Soln, Nag_Iter, Nag_Soln_Iter, Nag_Soln_Iter_Full};
  const int mprnlevels[]={Nag_NoPrint, Nag_Iter};

  if (!options)
    return;

#ifdef DONOTVERIFY
  // turn off the verification of derivatives, they are computed via
  // automatic differentiation so they must be exact
  options->verify_grad=Nag_NoCheck;
#endif

  if (param.tol_feas>0)
    options->major_feas_tol=param.tol_feas;

  if (param.min_feas>0)
    options->minor_feas_tol=param.min_feas;

  if (param.tol_prec>0)
    options->f_prec=param.tol_prec;

  if (param.tol_optim>0)
    options->major_opt_tol=param.tol_optim;

  if (param.maxit>=0)
    options->major_iter_lim=param.maxit;

  if (param.minit>=0)
    options->minor_iter_lim=param.minit;

  if (param.minittotal>0)
    options->iter_lim=param.minittotal;

  if (param.printlev>=0 && param.printlev<=4)
    options->print_level=prnlevels[param.printlev];
  else if (param.printlev>4)
    options->print_level=prnlevels[4];

  if (param.mprintlev>=0 && param.mprintlev<=1)
    options->minor_print_level=mprnlevels[param.mprintlev];
  else if (param.mprintlev>1)
    options->minor_print_level=mprnlevels[1];

  if (param.superbasics>0)
    options->max_sb=param.superbasics;

  if (param.hesupdates>=0)
    options->hess_update=param.hesupdates;

}

// function to evaluate constraints by calling AMPL functions
// expects in user allocated 2*N double, iuser[1] and further - lookuptable
static void NAG_CALL confun(Integer ncnln, Integer njnln, Integer nnzjac, const double x[], double conf[], double conjac[], Nag_Comm *comm) {

  double *xlong=comm->user;
  double *vtmp=comm->user+n_var;
  Integer *lookuptable=comm->iuser+1;
  Integer i,j,k;
  cgrad *cg;
  fint nerror=0;

  if (!comm)
    return;
  if (!x || !conf || !conjac || !comm->user || !comm->iuser) {
    comm->flag=-2;
    return;
  }

  //copy x->xlong with zeros where the variable is linear
  for (i=0; i<njnln; i++)
    xlong[i]=x[i];
  for (;i<n_var;i++)
    xlong[i]=0.0;

  if (comm->flag==0 || comm->flag==2) {  // function value
    for (i=0;i<ncnln;i++) {
      conf[i]=conival(i,xlong,&nerror);
      if (nerror) {
        comm->flag=-1;  // cannot be evaluated
        return;
      }
    }
  }
  if (comm->flag==1 || comm->flag==2) {  // derivative
    k=0; // counter of elements
    for (i=0; i<ncnln; i++) {
      congrd(i,xlong,vtmp,&nerror);
      if (nerror) {
        comm->flag=-1;  // cannot be evaluated
        return;
      }
      for (j=0,cg=Cgrad[i];cg;cg=cg->next) {
        if (cg->varno<njnln)
          conjac[lookuptable[k++]]=vtmp[j++]; 
          //fjac[k++]=vtmp[j++];   // ??? really like that? not ... store by columns
        else
          break;
      }
    }
  }
}

// function to evaluate value + derivative of the objective function by AMPL
// expects allocated 2*N double in ruser; iuser[0]=nobj
static void NAG_CALL objfun(Integer nonln, const double x[], double *objf, double objgrad[], Nag_Comm *comm) {

  int i;
  double *xlong=comm->user, *vtmp=comm->user+n_var;
  int nobj;
  fint nerror=0;

  if (!comm)
    return;
  if (!x || !objf || !objgrad || !comm->user || !comm->iuser) {
    comm->flag=-2;
    return;
  }

  nobj=comm->iuser[0];

  // copy x to the full vector, zeros anywhere outside nonlin vars
  for (i=0; i<nonln; i++)
    xlong[i]=x[i];
  for (;i<n_var;i++)
    xlong[i]=0.0;

  // evaluate objective function value
  if (comm->flag==0 || comm->flag==2) {
    if (nobj>=0) {
      *objf=objval(nobj,xlong,&nerror);
      if (nerror) {
        comm->flag=-1;  // cannot be evaluated
        return;
      }
    }
    else  // const. objective ... shouldn't be called but just for sure
      *objf=0;
  }

  // evaluate gradient of the objective
  if (comm->flag==1 || comm->flag==2) {
    if (nobj>=0) {
      objgrd(nobj, xlong, vtmp, &nerror);  //vtmp is not compressed
      if (nerror) {
        comm->flag=-1;  // cannot be evaluated
        return;
      }
      for (i=0;i<nonln;i++)
        objgrad[i]=vtmp[i];
    }
    else {       // const. objective
      for(i=0;i<nonln;i++)
        objgrad[i]=0.0;
    }
  }
}

int main(int argc, char *argv[]) {

  FILE *nlfile=0;
  char *stubname;
  char *progname=argv[0];
  int nobj=0;   // number of objective to use, -1 for constant
  int errflag;
  int retflag=0;
  int i, neediobj;
  Integer where;
  Integer nnzjnln=0;        //# nonzeroes in nonlinear part of jacobian
  Integer *lookuptable=0;   // where in A store nonlin. part of Jacobian
  Integer *kanln=0;         // 'ka' for nonlinear part of Jacobian
  double coef;
  cgrad *cg;
  ograd *og;
  int ifail;

  // e04ugc vars
  Integer n, m, ncnln, nonln, njnln, iobj, nnz, *ha=0, *ka=0, ninf, *iuser=0;
  double *a=0, *bl=0, *bu=0, *xs=0, sinf, obj, *ruser=0;
  double bigbnd=1e20;
  Nag_E04_Opt options;
  Nag_Comm comm;
  NagError fail;


  // prepare our AMPL problem structure to compure function values and 1st deriv
  asl = ASL_alloc(ASL_read_fg);
  // find problem name (stub) & react on -=, -?
  stubname = getstub(&argv, &Oinfo);
  if (!stubname) {
    printf("%s\n",Oinfo.bsname);
    printf("Licence: %s\n", nag_licence_query() ? "valid" : "valid licence NOT found");
    /*  With NAG_DLL it creates a pop-up window instead of the text to stdout
     *  so let's not use it
    printf("\nBased on library:\n");
    nag_implementation_details();
    */
    printf("\n");
    printf("No test problem set, terminating.\n");
    printf("Call `%s -?` for help.\n",progname);
    retflag=0;
    goto cleaning;
  }

  return_nofile = 1;
  // load basic characteristics of the problem
  nlfile = jac0dim(stubname, (ftnlen)strlen(stubname));
  if (!nlfile) {
    printf("Can't open %s\nTerminating.\n",stubname);
    retflag=1;
    goto cleaning;
  }

  // is it a "normal" nonlinear continuous problem?
  if (n_var<=0 || nbv>0 || niv>0 || nlnc>0 || lnc>0) {
    printf("Discrete problems or problems with newtork constraints are not supported by this solver.\nTerminating.\n");
    retflag=1;
    goto cleaning;
  }
  if (n_obj<=0) {
    printf("Warning: There is no objective function, pretending to be constant.\n");
    nobj=-1;
  } else if (n_obj>1) {
    printf("Warning: More than one objetive function detected, using the first one.\n");
  }
  
  // AMPL uses their "own" Malloc and type real - only to allocate their bits
  // by allocating these we notify AMPL to fill the data in
  // there are many macros in asl.h which changes many elements of the
  // structure 'asl' such that they look as local variables, for example:
  //    #define n_var   asl->i.n_var_
  //    #define LUv     asl->i.LUv_
  // So if there is no local definition of the variable it is likely to be
  // a part of 'asl' structure
  LUv = (real *)Malloc(n_var*sizeof(real));
  Uvx = (real *)Malloc(n_var*sizeof(real));
  LUrhs = (real *)Malloc(n_con*sizeof(real));
  Urhsx = (real *)Malloc(n_con*sizeof(real));
  X0 = (real *)Malloc(n_var*sizeof(real));
  pi0 = (real *)Malloc(n_con*sizeof(real));
  if (!X0 || !LUv || ! Uvx || (n_con && (!LUrhs || !Urhsx || !pi0))) {
    printf("Memory allocation problem, the problem is too big to fit into the memory. Terminating.\n");
    retflag=1;
    goto cleaning;
  }
  
  // read the rest of the problem in
  if ((errflag=fg_read(nlfile, ASL_return_read_err))!=0) {
    printf("AMPL problem with reading the problem, err_code: %i\nTerminating\n",errflag);
    retflag=1;
    goto cleaning;
  }

  // read additional options and store them into global 'param'
  getopts(argv, &Oinfo);  
  // initialize nag options structure and set it accordingly
  nag_opt_init(&options);
  set_options(&options);

  asl->i.congrd_mode = 1;  
     // condensed mode of gradients of constrants (don't store constant zeros)

  // correct min/max: AMPL: objtype[]==0 -> min, ==1 ->max
  if (nobj==-1) {
    // ignore objective function, find feasible point
    // (there is no objective anyway)
    options.direction=Nag_FeasiblePoint;
  } else if (objtype[nobj]) {
    // minimization is by default --> change it
    options.direction=Nag_Maximize;
  }

  // E04UGC init
  n=n_var;
  m=n_con;   // might need +1 for 'free' IOBJ row
  iobj=0;
  if (nobj!=-1 && nlvo<n_var) {  
    // ok, there are some variables which aren't nonlinear
    // but are they involved in the objective function at all?
    for (neediobj=0,og=Ograd[nobj]; og; og=og->next)
      if(og->varno>=nlvo) {
        neediobj=1;
        break;
      }
    if (neediobj)
      iobj=++m;  // is 1-based, last row
  } 
  if (m==0) {
    m=1;
    iobj=-1;   // 'dummy' free iobj row
  }
  ncnln=nlc;   //# nonlin. constr.
  nonln=nlvo;  //# nonlinear objective variables
  njnln=nlvc;  //# nonlinear constraints variables
  // find the structure of A (Jacobian of constraints with optional IOBJ)
  ka = (Integer *) calloc(n+1, sizeof(Integer));
  kanln = (Integer *) calloc(nlvc+1, sizeof(Integer));
  if (!ka || !kanln) {
    printf("Memory allocation problem, terminating.\n");
    retflag=1;
    goto cleaning;
  }
  if (iobj==-1) {
    ka[0]=0;     // ka[] is 0-based
    for (i=1;i<=n;i++)
      ka[i]=1;
    nnz=1;
  }
  else {
    for (i=0;i<n_con;i++)
      for (cg=Cgrad[i];cg;cg=cg->next) {
        ka[cg->varno+1]++;
        if (i<nlc && cg->varno<nlvc)
          kanln[cg->varno+1]++; // count number of nonzeroes in nonlin jac. part
          //nnzjnln++;  // count number of nonzeroes in nonlin jac. part
      }
    if (iobj>0)
      for (og=Ograd[nobj];og;og=og->next)
        ka[og->varno+1]++;
    for (i=1;i<n;i++)   // ka[0]=0 anyway -> skip
      ka[i+1]+=ka[i];   // careful - stil 0-based
    for (i=1;i<nlvc;i++)
      kanln[i+1]+=kanln[i];
    nnz=ka[n];
    nnzjnln=kanln[nlvc];
  }
  ninf=0;
  sinf=0.0;
  obj=0.0;
  a = (double *) calloc(nnz, sizeof(double));
  ha = (Integer *) calloc(nnz, sizeof(Integer));
  bl = (double *) malloc((n+m)*sizeof(double));
  bu = (double *) malloc((n+m)*sizeof(double));
  xs = (double *) calloc(n+m,sizeof(double));
  iuser = (Integer *) malloc((nnzjnln+1)*sizeof(Integer));
  ruser = (double *) malloc(2*n*sizeof(double));

  if (!a || !ha || !bl || !bu || !xs || !iuser || !ruser) {
    printf("Memory allocation problem, terminating.\n");
    retflag=1;
    goto cleaning;
  }
  iuser[0]=nobj;
  lookuptable=iuser+1;

  if (iobj==-1) {
    // nnz=1 & ka[] is ready
    ha[0]=1;
  }
  else {
    nnzjnln=0;
    for (i=0;i<nlc;i++)  // nonlinear constraints come first
      for (cg=Cgrad[i];cg;cg=cg->next) {
        where=ka[cg->varno]++;
        ha[where]=i+1;       //1-based
        a[where]=cg->coef;
        if (cg->varno<nlvc)
          lookuptable[nnzjnln++]=kanln[cg->varno]++;
      }
    for (;i<n_con;i++)  // cover the rest - linear
      for (cg=Cgrad[i];cg;cg=cg->next) {
        where=ka[cg->varno]++;
        ha[where]=i+1;       //1-based
        a[where]=cg->coef;
      }
    if (iobj>0)   // linear part comes last
      for (og=Ograd[nobj];og;og=og->next) {
        if (og->varno<nlvo)
          coef=0;
        else
          coef=og->coef;
        where=ka[og->varno]++;
        ha[where]=iobj;       //last row
        a[where]=coef;
      }

    //repair ka[] (& keep it 0-based)
    for (i=n;i>0;i--)
      ka[i]=ka[i-1];
    ka[0]=0;
  }

  // copy bounds
  for (i=0; i<n;i++) {
    bl[i]=LUv[i];
    bu[i]=Uvx[i];
  }
  for (i=0; i<n_con; i++) {
    bl[i+n]=LUrhs[i];
    bu[i+n]=Urhsx[i];
  }
  if (iobj>0) {   // should be iobj==m, i.e., 1-based
    bl[iobj-1+n]=-bigbnd;
    bu[iobj-1+n]=bigbnd;
  }
  else if (iobj==-1) {  // then m=1
    bl[n]=-bigbnd;
    bu[n]=bigbnd;
  }
  // initial estimate
  for (i=0;i<n;i++)
    xs[i]=X0[i];

  comm.iuser=iuser;
  comm.user=ruser;
  INIT_FAIL(fail);

  // optimize
  nag_opt_nlp_sparse(confun, objfun, n, m, ncnln, nonln, njnln, iobj, nnz,
         a, ha, ka, bl, bu, xs, &ninf, &sinf, &obj, &comm, &options, &fail);

  // results check
  switch(fail.code) {
    case NE_NOERROR:
      // OK, solution found
      ifail=0;
      printf("Result OK\n");
      break;

    case NE_NONLIN_NOT_FEASIBLE:
    case NE_LIN_NOT_FEASIBLE:
      // infeasible problems
      ifail=1;
      printf("Error from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    case NE_MAYBE_UNBOUNDED:
    case NE_UNBOUNDED:
      // unbounded problems
      ifail=2;
      printf("Error from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    case NE_BAD_PARAM:
    case NE_INVALID_REAL_RANGE_E:
    case NE_INVALID_REAL_RANGE_EE:
      // wrong option settings
      ifail=3;
      printf("Error in option settings of nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    case NE_TOO_MANY_ITER:
    case NE_TOO_MANY_MAJOR_ITER:
    case NE_TOO_MANY_MINOR_ITER:
      // maxit
      ifail=4;
      printf("Error from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    case NE_NOT_REQUIRED_ACC:
      // suboptimal solution found, accuracy not achieved
      ifail=5;
      printf("Warning from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    case NE_SUPERBASICS_LIMIT:
      // superbasics too low, increase via opt. settings
      ifail=6;
      printf("Error from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    // all errors reported to AMPL as 'general error, consult the manual'
    case NE_BASIS_SINGULAR:
    case NE_BASIS_ILL_COND:
    case NE_INTERNAL_ERROR:
    case NE_OUT_OF_WORKSPACE:
    case NE_USER_STOP:
    case NE_ALLOC_FAIL:
    case NE_CANNOT_CALCULATE:
    case NE_NO_IMPROVE:
      ifail=7;
      printf("Error from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;

    // these errors shouldn't happen - my coding errors or something strange
    case NE_2_INT_ARG_CONS:
    case NE_2_INT_OPT_ARG_CONS:
    case NE_3_INT_ARG_CONS:
    case NE_ARRAY_CONS:
    case NE_BOUND:
    case NE_BOUND_EQ:
    case NE_BOUND_EQ_LCON:
    case NE_BOUND_EQ_NLCON:
    case NE_BOUND_LCON:
    case NE_BOUND_NLCON:
    case NE_CON_DERIV_ERRORS:
    case NE_OBJ_DERIV_ERRORS:
    case NE_DUPLICATE_ELEMENT:
    case NE_INT_ARG_LT:
    case NE_INT_ARRAY_1:
    case NE_INT_ARRAY_2:
    case NE_INT_OPT_ARG_GT:
    case NE_INT_OPT_ARG_LT:
    case NE_INVALID_INT_RANGE_1:
    case NE_INVALID_INT_RANGE_2:
    case NE_NAME_TOO_LONG:
    case NE_NOT_APPEND_FILE:
    case NE_NOT_CLOSE_FILE:
    case NE_OBJ_BOUND:
    case NE_OPT_NOT_INIT:
    case NE_STATE_VAL:
    default:
      // strange errors which shouldn't happen
      ifail=7;
      printf("ERROR from nag_opt_nlp_sparse (e04ugc).\n%s\n", fail.message);
      break;
  }

  // write results to AMPL
  for (i=0;i<n_var;i++)
    X0[i]=xs[i];
  if (options.lambda) {
    for (i=0;i<n_con;i++)
      pi0[i]=options.lambda[n+i];
  } else {
    for (i=0;i<n_con;i++)
      pi0[i]=0;
  }
  
  // report the solution back to AMPL
  write_sol( (ifail>=0 && ifail<7) ? end_msg[ifail] : end_msg[7],X0,pi0,&Oinfo);

  // free all the option settings
  INIT_FAIL(fail);
  nag_opt_free(&options, "all", &fail);

cleaning: ;  // something to clean?
  if (ka) free(ka);
  if (kanln) free(kanln);
  if (a) free(a);
  if (ha) free(ha);
  if (bl) free(bl);
  if (bu) free(bu);
  if (xs) free(xs);
  if (iuser) free(iuser);
  if (ruser) free(ruser);

  if (asl)
    ASL_free(&asl);
      
  return retflag;
}

